/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * brush.c
 * -------
 *
 * A widget used to draw on a drawable widget
 */

#include "../../../config.h"

#include "impulse/rows/brush_int.h"
#include "impulse/rows/drawable_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"

/*                            ======--------======                           */

static void ROWS_brush_setup(ROWS_widget   *widget,
			     ROWS_drawable  drawable);
static int  brush_ID   = -1;
static int  brush_type = -1;

static void ROWS_brush_destroy(_p_ROWS_brush_data **private);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_brush_create(ROWS_drawable drawable)
{
  ROWS_widget *wid;

  if (brush_ID   == -1)
    brush_ID   = ROWS_widget_next_identifier();

  wid = ROWS_brush_initialise(brush_ID,
			      drawable);

  return *wid;
}

ROWS_widget *ROWS_brush_initialise(int id,
				   ROWS_drawable drawable)
{
  ROWS_widget *widget;

  if (brush_ID   == -1)
    brush_ID   = ROWS_widget_next_identifier();
  if (brush_type == -1)
    brush_type = ROWS_widget_next_type();

  widget = ROWS_widget_initialise(id);
  ROWS_widget_set_type(widget, brush_ID, brush_type);
  ROWS_widget_set_destructor(widget,
			     brush_ID,
			     (ROWS_widget_destroy_func) ROWS_brush_destroy);

  ROWS_brush_setup(widget, drawable);

  return widget;
}

int ROWS_brush_next_identifier(void)
{
  if (brush_ID==-1)
    brush_ID = ROWS_widget_next_identifier();

  return brush_ID + 1;
}

ROWS_brush ROWS_brush_cast(void **widget)
{
  if (brush_ID   == -1)
    brush_ID   = ROWS_widget_next_identifier();
  if (brush_type == -1)
    brush_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, brush_ID, brush_type);
}

void ROWS_brush_setup(ROWS_widget   *widget,
		      ROWS_drawable  drawable)
{
  _p_ROWS_brush_data          **private;
  _p_ROWS_drawable_data       **draw;
  ROWS_drawable_abstract_func   func;
  ROWS_drawable_fparam          param;

  private  = (_p_ROWS_brush_data **) ROWS_widget_private(widget, brush_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_brush_data));

  (*private)->drawable       = drawable;
  (*private)->colour         = NULL;
  (*private)->use_private_gc = 0;
  (*private)->last_serial    = 0;

  draw = (_p_ROWS_drawable_data **) drawable;
  
  func = (*draw)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_CREATE];

  param.brush.brush = (ROWS_brush) private;
  if (func!=NULL)
    (func)(&param);
}

void ROWS_brush_destroy(_p_ROWS_brush_data **private)
{
  _p_ROWS_drawable_data       **drawable;
  ROWS_drawable_abstract_func   func;
  ROWS_drawable_fparam          param;

  drawable = (_p_ROWS_drawable_data **) (*private)->drawable;
  
  func = (*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DESTROY];

  param.brush.brush = (ROWS_brush) private;
  if (func!=NULL)
    (func)(&param);
  
  /* FIXME */
  /*
  if ((*private)->use_private_gc)
    XFreeGC((*drawable)->display,
	    (*private)->private_gc);
  */

  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions */

extern void ROWS_brush_set_gc(ROWS_brush brush)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_GC])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_use_private_gc(ROWS_brush brush)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_USE_PRIVATE_GC])
    (&param);
  
  ROWS_widget_unlock(brush);
}

/* Brush state functions */

extern void ROWS_brush_set_linewidth (ROWS_brush brush,
				      microns    line_width)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.linewidth.width = line_width;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_LINEWIDTH])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_set_dotpattern(ROWS_brush            brush,
				      ROWS_brush_dotpattern pattern,
				      microns               offset)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.dotpattern.dp     = pattern;
  param.brush.data.dotpattern.offset = offset;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_DOTPATTERN])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_set_fgcolour  (ROWS_brush  brush,
				      ROWS_colour colour)
{
  _p_ROWS_brush_data         **data;
  _p_ROWS_drawable_data      **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.colour.colour = colour;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_FGCOLOUR])
    (&param);

  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_set_bgcolour  (ROWS_brush  brush,
				      ROWS_colour colour)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.colour.colour = colour;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_BGCOLOUR])
    (&param);

  ROWS_widget_unlock(brush);
}

extern int  ROWS_brush_get_fgpixel(ROWS_brush brush)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");
  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_GET_FGPIXEL])
    (&param);

  ROWS_widget_unlock(brush);

  return param.brush.data.singleint.value;
}

/* Draw with this brush */

extern void ROWS_brush_draw_point    (ROWS_brush brush,
				      microns x, microns y)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.coordinate.x = x;
  param.brush.data.coordinate.y = y;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DRAW_POINT])
    (&param);

  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_draw_line     (ROWS_brush brush,
				      microns x1, microns y1,
				      microns x2, microns y2)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.line.x1 = x1;
  param.brush.data.line.y1 = y1;
  param.brush.data.line.x2 = x2;
  param.brush.data.line.y2 = y2;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DRAW_LINE])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_draw_rectangle(ROWS_brush brush,
				      microns x, microns y,
				      microns w, microns h)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.line.x1 = x;
  param.brush.data.line.y1 = y;
  param.brush.data.line.x2 = w;
  param.brush.data.line.y2 = h;

  param.brush.brush = brush;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DRAW_RECTANGLE])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_draw_arc      (ROWS_brush brush,
				      microns x, microns y,
				      microns w, microns h,
				      int angle1, int angle2)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.arc.x = x;
  param.brush.data.arc.y = y;
  param.brush.data.arc.w = w;
  param.brush.data.arc.h = h;
  param.brush.data.arc.angle1 = angle1;
  param.brush.data.arc.angle2 = angle2;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DRAW_ARC])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_fill_rectangle(ROWS_brush brush,
				      microns x, microns y,
				      microns w, microns h)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.line.x1 = x;
  param.brush.data.line.y1 = y;
  param.brush.data.line.x2 = w;
  param.brush.data.line.y2 = h;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_FILL_RECTANGLE])
    (&param);
  
  ROWS_widget_unlock(brush);
}

extern void ROWS_brush_fill_arc      (ROWS_brush brush,
				      microns x, microns y,
				      microns w, microns h,
				      int angle1, int angle2)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.arc.x = x;
  param.brush.data.arc.y = y;
  param.brush.data.arc.w = w;
  param.brush.data.arc.h = h;
  param.brush.data.arc.angle1 = angle1;
  param.brush.data.arc.angle2 = angle2;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_FILL_ARC])
    (&param);
  
  ROWS_widget_unlock(brush);
}

void ROWS_brush_draw_shape(ROWS_brush        brush,
			   ROWS_brush_shape *points,
			   int               nopoints)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;
  param.brush.data.shape.shape = points;
  param.brush.data.shape.nopoints = nopoints;
  
  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DRAW_SHAPE])
    (&param);

  ROWS_widget_unlock(brush);
}

/*                            ======--------======                           */

/* Font functions */

void ROWS_brush_set_font(ROWS_brush  brush,
			 char       *font)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.set_font.font_name = font;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_FONT])
    (&param);
  
  ROWS_widget_unlock(brush);
}

void ROWS_brush_set_font_size(ROWS_brush brush,
			      microns    sizex,
			      microns    sizey)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.set_font_size.sizex = sizex;
  param.brush.data.set_font_size.sizey = sizey;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_SET_FONT_SIZE])
    (&param);
  
  ROWS_widget_unlock(brush);
}


void ROWS_brush_font_paint(ROWS_brush brush,
			   impulse_unistring  text,
			   microns    x,
			   microns    y)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.font_paint.use_matrix = 0;
  param.brush.data.font_paint.text       = text;
  param.brush.data.font_paint.x          = x;
  param.brush.data.font_paint.y          = y;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_FONT_PAINT])
    (&param);
  
  ROWS_widget_unlock(brush);
}

void ROWS_brush_font_paint_matrix(ROWS_brush brush,
				  impulse_unistring  text,
				  microns    x,
				  microns    y,
				  float     *matrix)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.font_paint.use_matrix = 1;
  param.brush.data.font_paint.text       = text;
  param.brush.data.font_paint.x          = x;
  param.brush.data.font_paint.y          = y;
  param.brush.data.font_paint.matrix     = matrix;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_FONT_PAINT_MATRIX])
    (&param);
  
  ROWS_widget_unlock(brush);
}

void ROWS_brush_font_measure(ROWS_brush          brush,
			     impulse_unistring           text,
			     ROWS_brush_measure *result)
{
  _p_ROWS_brush_data    **data;
  _p_ROWS_drawable_data **drawable;
  ROWS_drawable_fparam    param;

  ROWS_widget_check_type(brush, brush_type, "ROWS_brush");

  ROWS_widget_lock(brush);

  data     = (_p_ROWS_brush_data **) brush;
  drawable = (_p_ROWS_drawable_data **) (*data)->drawable;

  param.brush.brush = brush;

  param.brush.data.font_measure.text   = text;
  param.brush.data.font_measure.result = result;

  ((*drawable)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_FONT_MEASURE])
    (&param);
  
  ROWS_widget_unlock(brush);
}
