/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * child.c
 * -------
 *
 * Class of widgets child in containers
 */

#include "../../../config.h"

#include "impulse/rows/child_int.h"

#include "impulse/util/memory.h"
#include "impulse/util/unistring.h"
#include "impulse/util/sparse.h"
#include "impulse/util/thread.h"

#if HAVE_GUI_XWIN
# include "impulse/rows/child_xwin.c"
#else
#error "Child windows not implemented on this platform"
#endif

/*                            ======--------======                           */

static void ROWS_child_setup(ROWS_widget   *widget,
			     ROWS_drawable  drawable);
       int  _p_ROWS_child_ID   = -1;
       int  _p_ROWS_child_type = -1;

static void ROWS_child_destroy(void **private);

static impulse_bool _p_ROWS_child_move(ROWS_child            widget,
				       ROWS_child_event_num  event_number,
				       ROWS_child_event     *event,
				       void                 *data);

/*                            ======--------======                           */

static impulse_sparse selected_children = NULL;
typedef struct _p_ROWS_child_select
{
  ROWS_child select;
} _p_ROWS_child_select;
#if USE_THREADS
static impulse_mutex set_state_mutex = NULL;
#endif

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_child_create(ROWS_drawable drawable)
{
  ROWS_widget *wid;

  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();

  wid = ROWS_child_initialise(_p_ROWS_child_ID,
			      drawable);

  return *wid;
}


ROWS_widget *ROWS_child_initialise(int id,
				   ROWS_drawable drawable)
{
  ROWS_widget *widget;

  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();
  if (_p_ROWS_child_type == -1)
    _p_ROWS_child_type = ROWS_widget_next_type();

  widget = ROWS_eventbox_initialise(id);
  ROWS_widget_set_type(widget, _p_ROWS_child_ID, _p_ROWS_child_type);
  ROWS_widget_set_destructor(widget, _p_ROWS_child_ID, ROWS_child_destroy);

  ROWS_child_setup(widget,
		   drawable);

  return widget;
}


int ROWS_child_next_identifier(void)
{
  if (_p_ROWS_child_ID==-1)
    _p_ROWS_child_ID = ROWS_eventbox_next_identifier();

  return _p_ROWS_child_ID + 1;
}


ROWS_child ROWS_child_cast(void **widget)
{
  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();
  if (_p_ROWS_child_type == -1)
    _p_ROWS_child_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, _p_ROWS_child_ID, _p_ROWS_child_type);
}


void ROWS_child_setup(ROWS_widget   *widget,
		      ROWS_drawable  drawable)
{
  _p_ROWS_child_data **priv;

  priv  = (_p_ROWS_child_data **) ROWS_widget_private(widget, _p_ROWS_child_ID);
  *priv = impulse_alloc(sizeof(_p_ROWS_child_data));

  if (!selected_children)
  {
    impulse_mutex_global_lock;
    if (!selected_children)
    {
#if USE_THREADS
      set_state_mutex = impulse_mutex_create_recursive();
#endif
      selected_children = impulse_sparse_create();
    }
    impulse_mutex_global_unlock;
  }

  ROWS_eventbox_register_widget(ROWS_eventbox_cast(*widget),
				_p_ROWS_child_ID,
				ROWS_CHILD_EVENT_MAX);

  (*priv)->draw = drawable;

  (*priv)->active   = ROWS_STYLE_NORMAL;
  (*priv)->border.x = 1;
  (*priv)->border.y = 1;
  (*priv)->offset.x = 0;
  (*priv)->offset.y = 0;

  (*priv)->realised   = 0;
  (*priv)->actualised = 0;
  (*priv)->visible = 0;
  (*priv)->activation = ROWS_STYLE_NORMAL;

  (*priv)->container = NULL;

  (*priv)->state        = 0;//ROWS_CHILD_UNSELECTED;
  (*priv)->select_group = 0;
  
  if (drawable!=NULL)
    {
      (*priv)->style = ROWS_drawable_get_style(drawable);
      ROWS_style_reference((*priv)->style);
    }
  else
    (*priv)->style = NULL;
      
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_MOVE,
			    _p_ROWS_child_move,
			    NULL);

  _p_ROWS_child_int_setup(widget, drawable);
}


void ROWS_child_destroy(void **data)
{
  _p_ROWS_child_data **private;

  private  = (_p_ROWS_child_data **) data;

  if ((*private)->select_group != 0)
    {
      /*
       * The observant among you will note that set_state_mutex must
       * be initialised for this widget to have a grouping
       */
#if USE_THREADS
      impulse_mutex_lock(set_state_mutex);
#endif
      if ((*private)->select_group != 0)
	{
	  impulse_sparse_add(selected_children,
			     (*private)->select_group,
			     NULL);
	}
#if USE_THREADS
      impulse_mutex_unlock(set_state_mutex);
#endif
    }
  
  _p_ROWS_child_int_destroy(private);
  ROWS_style_dereference((*private)->style);

  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions */

/* Locking functions */


void ROWS_child_lock(ROWS_child child)
{
#if USE_THREADS
  _p_ROWS_child_data **private;
  ROWS_container       lock_container;

  if (child == NULL)
    return;

  private = (_p_ROWS_child_data **) ROWS_child_cast(child);

  /* We lock the container first, then the child */
  do {
    lock_container = (*private)->container;
    ROWS_child_lock(lock_container);
    (*private)->lock_container = lock_container;

    ROWS_widget_lock(child);

    /* If we've been moved around since we got the lock, deal with it */
    if (lock_container != (*private)->container)
      {
	ROWS_child_unlock(lock_container);
	ROWS_widget_unlock(child);
      }
  } while (lock_container != (*private)->container);

#endif
}


void ROWS_child_unlock(ROWS_child child)
{
#if USE_THREADS
  _p_ROWS_child_data **private;
  
  if (child == NULL)
    return;

  private = (_p_ROWS_child_data **) ROWS_child_cast(child);

  ROWS_child_unlock((*private)->lock_container);
  ROWS_widget_unlock(child);
#endif
}


/* Event functions */

void ROWS_child_event_register(ROWS_child            child,
			       ROWS_child_event_num  event,
			       ROWS_child_func       func,
			       void                 *data)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  ROWS_eventbox_register_handler(ROWS_eventbox_cast(child),
				 _p_ROWS_child_ID,
				 (ROWS_eventbox_func) func,
				 event,
				 data);
}


void ROWS_child_event_deregister(ROWS_child            child,
				 ROWS_child_event_num  event,
				 ROWS_child_func       func,
				 void                 *data)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  ROWS_eventbox_deregister_handler(ROWS_eventbox_cast(child),
				   _p_ROWS_child_ID,
				   (ROWS_eventbox_func) func,
				   event,
				   data);
}


void ROWS_child_event_send(ROWS_child            child,
			   ROWS_child_event_num  event,
			   ROWS_child_event     *data)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  ROWS_child_lock(child);
  
  ROWS_eventbox_send(ROWS_eventbox_cast(child),
		     _p_ROWS_child_ID,
		     child,
		     event,
		     data);
  
  ROWS_child_unlock(child);
}


impulse_bool ROWS_child_event_is_handled(ROWS_child child,
					 ROWS_child_event_num type)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");
  
  return ROWS_eventbox_is_handled(ROWS_eventbox_cast(child),
				  _p_ROWS_child_ID,
				  type);
}


/* Geometry functions */

void ROWS_child_set_size(ROWS_child  child,
			 ROWS_point *size)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  if (ROWS_eventbox_is_handled(ROWS_eventbox_cast(child),
			       _p_ROWS_child_ID,
			       ROWS_CHILD_PRESIZE))
  {
    ROWS_child_event event;
  
    event.size.size = *size;
    ROWS_child_event_send(child,
			  ROWS_CHILD_PRESIZE,
			  &event);
  }
  else
  {
    ROWS_child_event event;
  
    ROWS_box_set_physical(ROWS_box_cast(child),
			  size);
    ROWS_box_set_virtual(ROWS_box_cast(child),
			 size);
  
    event.size.size = *size;
    ROWS_child_event_send(child,
			  ROWS_CHILD_SIZE,
			  &event);
  }
}


ROWS_point *ROWS_child_get_size(ROWS_child child)
{
  return ROWS_box_get_physical(ROWS_box_cast(child));
}


void ROWS_child_set_border(ROWS_child  child,
			   ROWS_point *border)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  (*private)->border = *border;
  ROWS_widget_unlock(child);
}


ROWS_point *ROWS_child_get_border(ROWS_child child)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return &((*private)->border);
}


void ROWS_child_set_origin(ROWS_child  child,
			   ROWS_point *origin)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  if (ROWS_eventbox_is_handled(ROWS_eventbox_cast(child),
			       _p_ROWS_child_ID,
			       ROWS_CHILD_PREMOVE))
  {
    ROWS_child_event event;
  
    event.move.position = *origin;
    ROWS_child_event_send(child,
			  ROWS_CHILD_PREMOVE,
			  &event);
  }
  else
  {
    ROWS_child_event event;
  
    ROWS_box_set_position(ROWS_box_cast(child),
			  origin);
    
    event.move.position = *origin;
    ROWS_child_event_send(child,
			  ROWS_CHILD_MOVE,
			  &event);
  }
}


ROWS_point *ROWS_child_get_origin(ROWS_child child)
{
  return ROWS_box_get_position(ROWS_box_cast(child));
}


ROWS_point *ROWS_child_get_offset(ROWS_child child)
{
  _p_ROWS_child_data **priv;

  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  return &((*priv)->offset);  
}


/* Display/hiding functions */

void ROWS_child_set_visible(ROWS_child child,
			    impulse_bool visible)
{
  _p_ROWS_child_data **priv;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  if (visible == (*priv)->visible)
  {
    ROWS_child_unlock(child);
    return;
  }
  
  if ((*priv)->realised == 0)
    ROWS_child_set_realised(child, 1);

  _p_ROWS_child_int_set_visible(child,
				visible);

  ROWS_child_unlock(child);
}


impulse_bool ROWS_child_is_visible(ROWS_child child)
{
  _p_ROWS_child_data **priv;
  impulse_bool visible;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  visible = (*priv)->visible;

  ROWS_child_unlock(child);
  
  return visible;
}


void ROWS_child_set_realised(ROWS_child child,
			     impulse_bool realised)
{
  _p_ROWS_child_data **priv;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  if ((*priv)->realised == realised)
  {
    ROWS_child_unlock(child);
    return;
  }

  _p_ROWS_child_int_set_realised(child,
				 realised);

  ROWS_child_unlock(child);
}


impulse_bool ROWS_child_is_realised(ROWS_child child)
{
  _p_ROWS_child_data **priv;
  impulse_bool realised;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  realised = (*priv)->realised;

  ROWS_child_unlock(child);
  
  return realised;
}


void ROWS_child_set_actualised(ROWS_child child,
			       impulse_bool actual)
{
  _p_ROWS_child_data **priv;
  int realise, visible;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  if ((*priv)->actualised == actual)
  {
    ROWS_child_unlock(child);
    return;
  }

  realise = (*priv)->realised;
  visible = (*priv)->visible;
  ROWS_child_set_realised(child, 0);
  _p_ROWS_child_int_set_actualised(child,
				   actual);
  ROWS_child_set_realised(child, realise);
  ROWS_child_set_visible(child, visible);

  ROWS_child_unlock(child);
}


impulse_bool ROWS_child_is_actualised(ROWS_child child)
{
  _p_ROWS_child_data **priv;
  impulse_bool actualised;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;

  actualised = (*priv)->actualised;

  ROWS_child_unlock(child);
  
  return actualised;
}

/* Set-up ready to draw functions */


void ROWS_child_set_container(ROWS_child child,
			      ROWS_container container)
{
  _p_ROWS_child_data **priv;
  
  ROWS_widget_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");
  
  priv = (_p_ROWS_child_data **)child;
  (*priv)->container = container;

  ROWS_child_set_drawable(child, container ? ROWS_child_get_drawable(ROWS_child_cast(container)) : NULL);
  
  _p_ROWS_child_int_container(child, container);
  ROWS_widget_unlock(child);
}


void ROWS_child_set_drawable(ROWS_child child,
			     ROWS_drawable drawable)
{
  _p_ROWS_child_data **priv;

  ROWS_widget_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **)child;

  if ((*priv)->draw)
  {
    ROWS_style_dereference((*priv)->style);
    (*priv)->style = NULL;
    (*priv)->draw = NULL;
  }
  
  if (drawable)
  {
    (*priv)->draw = drawable;
    (*priv)->style = ROWS_drawable_get_style(drawable);
    ROWS_style_reference((*priv)->style);
  }

  ROWS_widget_unlock(child);
}


ROWS_drawable ROWS_child_get_drawable(ROWS_child child)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return (*private)->draw;
}


ROWS_container ROWS_child_get_container(ROWS_child child)
{
  _p_ROWS_child_data **private;

  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return (*private)->container;
}


ROWS_style ROWS_child_get_style(ROWS_child child)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return (*private)->style;
}


void ROWS_child_update(ROWS_child child)
{
  _p_ROWS_child_data **private;
  ROWS_rect rect;

  ROWS_child_lock(child);

  private = (_p_ROWS_child_data **) child;

  if ((*private)->draw == NULL)
    {
      ROWS_child_unlock(child);
      return;
    }

  ROWS_widget_lock((*private)->draw);

  _p_ROWS_child_int_update(child);

  rect.size = *ROWS_child_get_size(child);
  rect.pos = *ROWS_child_get_offset(child);

  ROWS_drawable_clip_push((*private)->draw);
  ROWS_drawable_clip_clear((*private)->draw);
  ROWS_drawable_clip_union((*private)->draw, &rect);
  ROWS_drawable_clip_set((*private)->draw);
  
  ROWS_drawable_clear_area((*private)->draw, 
			   rect.pos.x, rect.pos.y,
			   rect.size.x, rect.size.y);
  ROWS_drawable_clip_pop((*private)->draw);
  
  ROWS_drawable_sync((*private)->draw);
  ROWS_widget_unlock((*private)->draw);

  ROWS_child_unlock(child);
}


/* Misc stuff */

void ROWS_child_set_focus(ROWS_child child,
			  impulse_bool focused)
{
  /* Fix me: implement me */
}


impulse_bool ROWS_child_has_focus(ROWS_child child)
{
  /* Fix me: implement me */
  return 0;
}


/* Activation state functions */

void ROWS_child_set_activation(ROWS_child             child,
			       ROWS_style_active_type type)
{
  _p_ROWS_child_data **private;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  (*private)->active = type;
  ROWS_child_update(child);
  
  ROWS_child_unlock(child);
}

ROWS_style_active_type ROWS_child_get_activation(ROWS_child child)
{
  _p_ROWS_child_data     **private;
  ROWS_style_active_type   style;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  style = (*private)->active;
  
  ROWS_child_unlock(child);

  return style;
}

/* Realisation functions */


/* Internal event functions */

     
static impulse_bool _p_ROWS_child_move(ROWS_child            widget,
				       ROWS_child_event_num  event_number,
				       ROWS_child_event     *event,
				       void                 *data)
{
  _p_ROWS_child_data **priv = (_p_ROWS_child_data **)widget;
  
  if ((*priv)->actualised || !(*priv)->container)
  {
    (*priv)->offset.x = 0;
    (*priv)->offset.y = 0;
  }
  else
  {
    ROWS_point *offset = ROWS_child_get_offset(ROWS_child_cast((*priv)->container));
    (*priv)->offset.x = offset->x + event->move.position.x;
    (*priv)->offset.y = offset->y + event->move.position.y;
  }
  return 0;
}


/*                            ======--------======                           */

void ROWS_child_set_state(ROWS_child       child,
			  ROWS_child_state state,
			  impulse_int            grouping)
{
  _p_ROWS_child_data **private;

  ROWS_widget_lock(child);
#if USE_THREADS
  impulse_mutex_lock(set_state_mutex);
#endif
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  if (grouping != 0)
    {
      _p_ROWS_child_select *old_select, *new_select;

      old_select = impulse_sparse_recover(selected_children,
					  grouping);
      if (old_select != NULL)
	if (old_select->select != child && old_select->select != NULL)
	  {
	    ROWS_widget_unlock(child);

//	    ROWS_child_set_state(old_select->select,
//				 ROWS_CHILD_UNSELECTED,
//				 0);
	    
	    ROWS_widget_lock(child);
	  }

      new_select = impulse_alloc(sizeof(_p_ROWS_child_select));
      new_select->select = child;

      impulse_sparse_add(selected_children,
			 grouping,
			 new_select);
    }

  (*private)->state        = state;
  (*private)->select_group = grouping;
#if USE_THREADS
  impulse_mutex_unlock(set_state_mutex);
#endif
  ROWS_widget_unlock(child);  
}


/* End of file */
