/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * container.c
 * -----------
 *
 * Class of widgets containing containers
 */

#include "../../../config.h"

#include "impulse/rows/container_int.h"
#include "impulse/rows/child_int.h"
#include "impulse/util/memory.h"

/*                            ======--------======                           */

static void ROWS_container_setup(ROWS_widget   *widget);
static int  container_ID   = -1;
static int  container_type = -1;

static void ROWS_container_destroy(void **private);

/* Contained events */
static impulse_bool  _p_ROWS_container_realise(ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                     *data);
static impulse_bool  _p_ROWS_container_redraw (ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                     *data);
static impulse_bool  _p_ROWS_container_resize (ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                     *data);
static impulse_bool  _p_ROWS_container_button (ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                     *data);
static impulse_bool  _p_ROWS_container_crossing(ROWS_child            widget,
						ROWS_child_event_num  event_number,
						ROWS_child_event     *event,
						void                     *data);
static impulse_bool  _p_ROWS_container_motion(ROWS_child            widget,
					      ROWS_child_event_num  event_number,
					      ROWS_child_event     *event,
					      void                     *data);


/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_container_create(ROWS_drawable drawable)
{
  ROWS_widget *wid;

  if (container_ID   == -1)
    container_ID   = ROWS_child_next_identifier();

  wid = ROWS_container_initialise(container_ID,
				  drawable);

  return *wid;
}


ROWS_widget *ROWS_container_initialise(int id,
				       ROWS_drawable drawable)
{
  ROWS_widget *widget;

  if (container_ID   == -1)
    container_ID   = ROWS_child_next_identifier();
  if (container_type == -1)
    container_type = ROWS_widget_next_type();

  widget = ROWS_child_initialise(id, drawable);
  
  ROWS_widget_set_type(widget, container_ID, container_type);
  ROWS_widget_set_destructor(widget, container_ID, ROWS_container_destroy);

  ROWS_container_setup(widget);

  return widget;
}


int ROWS_container_next_identifier(void)
{
  if (container_ID==-1)
    container_ID = ROWS_child_next_identifier();

  return container_ID + 1;
}


ROWS_container ROWS_container_cast(void **widget)
{
  if (container_ID   == -1)
    container_ID   = ROWS_child_next_identifier();
  if (container_type == -1)
    container_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, container_ID, container_type);
}


void ROWS_container_setup(ROWS_widget *widget)
{
  _p_ROWS_container_data **priv;
  ROWS_child           child;
  ROWS_container           container;

  priv = (_p_ROWS_container_data **) ROWS_widget_private(widget, container_ID);
  *priv = impulse_alloc(sizeof(_p_ROWS_container_data));

  (*priv)->children = NULL;
  (*priv)->mouse_contained = NULL;
  (*priv)->mouse_grabbed = NULL;
  (*priv)->mouse_grab = 0;
  (*priv)->mouse_ignore = 0;

  child = ROWS_child_cast(*widget);
  container = ROWS_container_cast(*widget);

  ROWS_child_event_register(child,
			    ROWS_CHILD_REALISATION,
			    _p_ROWS_container_realise,
			    NULL);
  ROWS_child_event_register(child,
			    ROWS_CHILD_EXPOSURE,
			    _p_ROWS_container_redraw,
			    NULL);
  
  ROWS_child_event_register(child,
			    ROWS_CHILD_SIZE,
			    _p_ROWS_container_resize,
			    NULL);
  
  ROWS_child_event_register(child,
			    ROWS_CHILD_BUTTON,
			    _p_ROWS_container_button,
			    NULL);
  ROWS_child_event_register(child,
			    ROWS_CHILD_CROSSING,
			    _p_ROWS_container_crossing,
			    NULL);
  ROWS_child_event_register(child,
			    ROWS_CHILD_MOTION,
			    _p_ROWS_container_motion,
			    NULL);
}


void ROWS_container_destroy(void **data)
{
  _p_ROWS_container_data **priv;

  priv = (_p_ROWS_container_data **) data;

  impulse_free(*priv);
}

/*                            ======--------======                           */

/* Responses to various events */

static impulse_bool _p_ROWS_container_realise(ROWS_child            widget,
					      ROWS_child_event_num  event_number,
					      ROWS_child_event     *event,
					      void                     *data)
/*
 * Upon realisation, give all the child widgets a drawable and then
 * realise them too.
 */
{
  ROWS_container             container;
  _p_ROWS_container_data   **priv;
  ROWS_container_child  *next;

  container = ROWS_container_cast(widget);
  priv = (_p_ROWS_container_data **) container;

  next = (*priv)->children;
  
  while (next)
  {
    if (!ROWS_child_get_drawable(next->child))
      ROWS_child_set_drawable(next->child,
			      ROWS_child_get_drawable(widget));
    
    ROWS_child_set_realised(next->child,1);
    next = next->next;
  }

  return 0; /* Allow control to continue */
}


static impulse_bool _p_ROWS_container_redraw(ROWS_child            widget,
					     ROWS_child_event_num  event_number,
					     ROWS_child_event     *event,
					     void                     *data)
{
  _p_ROWS_container_data **priv;
  ROWS_container_child    *next;
  
  priv = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  next = (*priv)->children;
  while (next)
  {
    _p_ROWS_child_data **child = (_p_ROWS_child_data **) ROWS_child_cast(next->child);

    if ((*child)->visible &&
	!(*child)->actualised)
    {
      ROWS_rect rect;
      ROWS_point *origin;
      
      origin = ROWS_child_get_origin((ROWS_child)child);
      rect.pos = *ROWS_child_get_offset((ROWS_child)child);
      rect.size = *ROWS_child_get_size((ROWS_child)child);

#if 0
      /* 
       * This bit of code doesn't currently work. There are issues
       * with rounding between pixels and microns.
       * 
       * The basic problem is that the child offset, size, and origin
       * are calculated exactly in microns, whereas the clip mask is
       * calculated from microns, to pixels and back to microns, loosing
       * some precision. Feck.
       */
      printf("Child at %f,%f, Clipping to %f,%f at %f,%f\n", 
	     origin->x,
	     origin->y, 
	     event->exposure.clip.size.x, 
	     event->exposure.clip.size.y,
	     event->exposure.clip.pos.x, 
	     event->exposure.clip.pos.y);
      printf("Child at %d,%d, Clipping to %d,%d at %d,%d\n",
	     MICRONS_TO_PIXEL(origin->x),
	     MICRONS_TO_PIXEL(origin->y), 
	     MICRONS_TO_PIXEL(event->exposure.clip.size.x), 
	     MICRONS_TO_PIXEL(event->exposure.clip.size.y),
	     MICRONS_TO_PIXEL(event->exposure.clip.pos.x), 
	     MICRONS_TO_PIXEL(event->exposure.clip.pos.y));
      if (origin->x >= event->exposure.clip.pos.x)
	printf("Stage 1\n");
      if (origin->x < (event->exposure.clip.pos.x + event->exposure.clip.size.x))
	printf("Stage 2\n");
      if (origin->y >= event->exposure.clip.pos.y)
	printf("Stage 3\n");
      if (origin->y < (event->exposure.clip.pos.y + event->exposure.clip.size.y))
	printf("Stage 4\n");
      if ((origin->x >= event->exposure.clip.pos.x &&
	   origin->x < (event->exposure.clip.pos.x + event->exposure.clip.size.x)) &&
	  (origin->y >= event->exposure.clip.pos.y &&
	   origin->y < (event->exposure.clip.pos.y + event->exposure.clip.size.y)))
#endif
      {
//	ROWS_drawable_clip_push(drawable);
//	ROWS_drawable_clip_intersect(drawable, &rect);
//	ROWS_drawable_clip_set(drawable);
	
	ROWS_child_event_send(ROWS_child_cast(next->child),
			      ROWS_CHILD_EXPOSURE,
			      event);
	
//	ROWS_drawable_clip_pop(drawable);
      }
    }
    next = next->next;
  }

  return 0; /* Allow control to continue */
}


static impulse_bool _p_ROWS_container_button(ROWS_child            widget,
					     ROWS_child_event_num  event_number,
					     ROWS_child_event     *event,
					     void                 *data)
{
  _p_ROWS_container_data **priv;
  priv = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  if ((*priv)->mouse_ignore)
  {
    /* 
     * If all buttons have been released we can stopp ignoring
     * events
     */
    if (!event->button.pressed &&
	!(event->button.state & ROWS_STATE_BUTTONS))
      (*priv)->mouse_ignore = 0;
    return 0;
  }

  /*
   * Time to instigate a mouse grab.
   */
  if ((event->button.state & ROWS_STATE_BUTTONS) &&
      !(*priv)->mouse_grabbed)
  {
    (*priv)->mouse_grab = 1;
    /*
     * NB. mouse_grabbed will be NULL if mouse is not
     * in a child window (ie. the container has the grab)
     */
    (*priv)->mouse_grabbed = (*priv)->mouse_contained;
  }

  /*
   * Forward the button event onto the child with
   * the mouse grab
   */
  if ((*priv)->mouse_grabbed)
  {
    ROWS_point *adjust = ROWS_box_get_position(ROWS_box_cast((*priv)->mouse_grabbed));
    ROWS_child_event ev = *event;
    ev.button.position.x -= adjust->x;
    ev.button.position.y -= adjust->y;
    
    ROWS_child_event_send((*priv)->mouse_grabbed,
			  ROWS_CHILD_BUTTON,
			  &ev);
  }

  /* 
   * Time to release the grab
   */
  if (!(event->button.state & ROWS_STATE_BUTTONS))
  {
    (*priv)->mouse_grab = 0;
    (*priv)->mouse_grabbed = NULL;
  }
  return 0;
}


static impulse_bool _p_ROWS_container_crossing(ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                 *data)
{
  _p_ROWS_container_data **priv;
  priv = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  if (!event->crossing.entered)
  {
    /*
     * If mouse exits the container, then we also send exit
     * message to any child containing the mouse
     */
    if ((*priv)->mouse_contained)
    {
      ROWS_point *adjust = ROWS_box_get_position(ROWS_box_cast((*priv)->mouse_contained));
      ROWS_child_event ev = *event;
      
      ev.crossing.position.x -= adjust->x;
      ev.crossing.position.y -= adjust->y;
      
      ROWS_child_event_send((*priv)->mouse_contained,
			    ROWS_CHILD_CROSSING,
			    &ev);
      (*priv)->mouse_contained = NULL;
    }
  }
  else
  {
    ROWS_container_child *next = (*priv)->children;
    if ((*priv)->mouse_contained != NULL)
    {
      printf("*** FECK - mouse container is not null\n");
      (*priv)->mouse_contained = NULL;
    }
    
    /*
     * Mouse has entered the container, so its time to see if
     * it is in any children
     */
    while (next)
    {
      _p_ROWS_child_data **child = (_p_ROWS_child_data **) ROWS_child_cast(next->child);
      ROWS_box box = ROWS_box_cast(next->child);
      
      /*
       * NB. We do not consider actualised children, cos we get
       * real enter/exit events for them from the GUI
       */
      if ((*child)->visible &&
	  !(*child)->actualised &&
	  ROWS_box_contains(box,
			    &event->crossing.position))
      {
	if (!(*priv)->mouse_grab ||
	    ((*priv)->mouse_grabbed == ROWS_child_cast(next->child)))
	{
	  ROWS_point *adjust = ROWS_box_get_position(box);
	  ROWS_child_event ev = *event;

	  (*priv)->mouse_contained = ROWS_child_cast(next->child);
	  
	  ev.crossing.position.x -= adjust->x;
	  ev.crossing.position.y -= adjust->y;
	  
	  ROWS_child_event_send((*priv)->mouse_contained,
				ROWS_CHILD_CROSSING,
				&ev);
	}
	break;
      }
      next = next->next;
    }
  }

  return 0;
}


static impulse_bool _p_ROWS_container_motion(ROWS_child            widget,
					     ROWS_child_event_num  event_number,
					     ROWS_child_event     *event,
					     void                 *data)
{
  _p_ROWS_container_data **priv;
  ROWS_container_child *next;
  ROWS_child contained = NULL;

  priv = (_p_ROWS_container_data **) ROWS_container_cast(widget);


  /* 
   * First off, check to see which child contains the mouse event
   */
  next = (*priv)->children;
  while (next && !contained)
  {
    _p_ROWS_child_data **child = (_p_ROWS_child_data **) ROWS_child_cast(next->child);
    
    if ((*child)->visible &&
	!(*child)->actualised &&
	ROWS_box_contains(ROWS_box_cast(next->child),
			  &event->motion.position))
      contained = ROWS_child_cast(next->child);
    next = next->next;
  }

  /* 
   * Now check to see if we need to generate any enter/exit events
   */
  if (contained != (*priv)->mouse_contained)
  {
    if ((*priv)->mouse_contained)
    {
      ROWS_child_event ev;
      ROWS_point *adjust = ROWS_box_get_position(ROWS_box_cast((*priv)->mouse_contained));
      ev.crossing.position.x = event->motion.position.x - adjust->x;
      ev.crossing.position.y = event->motion.position.y - adjust->y;
      ev.crossing.state = event->motion.state;
      ev.crossing.entered = 0;
      ROWS_child_event_send((*priv)->mouse_contained,
			    ROWS_CHILD_CROSSING,
			    &ev);
    }
    
    if (!(*priv)->mouse_grab ||
	((*priv)->mouse_grabbed == contained))
      (*priv)->mouse_contained = contained;
    else
      (*priv)->mouse_contained = NULL;
    if ((*priv)->mouse_contained)
    {
      ROWS_child_event ev;
      ROWS_point *adjust = ROWS_box_get_position(ROWS_box_cast((*priv)->mouse_contained));
      ev.crossing.position.x = event->motion.position.x - adjust->x;
      ev.crossing.position.y = event->motion.position.y - adjust->y;
      ev.crossing.state = event->motion.state;
      ev.crossing.entered = 1;
      ROWS_child_event_send((*priv)->mouse_contained,
			    ROWS_CHILD_CROSSING,
			    &ev);
    }
  }
  
  /*
   * If the mouse is grabbed, then send the move event to 
   * the child with the grab
   */
  contained = NULL;
  if ((*priv)->mouse_grabbed)
    contained = (*priv)->mouse_grabbed;
  else if ((*priv)->mouse_contained)
    contained = (*priv)->mouse_contained;

  if (contained)
  {
    ROWS_point *adjust = ROWS_box_get_position(ROWS_box_cast(contained));
    ROWS_child_event ev = *event;
    
    ev.motion.position.x -= adjust->x;
    ev.motion.position.y -= adjust->y;

    ROWS_child_event_send(contained,
			  ROWS_CHILD_MOTION,
			  &ev);
  }

  return 0;
}


static impulse_bool _p_ROWS_container_resize(ROWS_child            widget,
					     ROWS_child_event_num  event_number,
					     ROWS_child_event     *event,
					     void                 *data)
{
  /* We should probably invalid the mouse_child member
   * here ? */
  return 0;
}


/*                            ======--------======                           */

void ROWS_container_add_int(ROWS_container container,
			    ROWS_child child,
			    void *data)
{
  _p_ROWS_container_data   **priv;
  ROWS_container_child      *next;

  ROWS_child_lock(container);
  ROWS_widget_lock(child);
  ROWS_widget_check_type(container, container_type, "ROWS_container");
  
  priv    = (_p_ROWS_container_data **)container;

  next = impulse_alloc(sizeof(ROWS_container_child));
  next->next = (*priv)->children;
  next->child = child;
  next->data = data;

  (*priv)->children = next;

  ROWS_child_set_container(child, container);

  ROWS_widget_unlock(child);
  ROWS_child_unlock(container);
}


void *ROWS_container_remove_int(ROWS_container container,
			       ROWS_child child)
{
  _p_ROWS_container_data   **priv;
  ROWS_container_child      *next, *prev;
  void *data = NULL;

  ROWS_child_lock(container);
  ROWS_widget_check_type(container, container_type, "ROWS_container");
  
  priv = (_p_ROWS_container_data **)container;

  next = (*priv)->children;
  prev = NULL;
  while (next)
  {
    if (next->child == child)
    {
      if (prev)
	prev->next = next->next;
      else
	(*priv)->children = next->next;
      data = next->data;
      impulse_free(next);
      break;
    }
    prev = next;
    next = next->next;
  }

  ROWS_child_set_container(child,
			   NULL);

  ROWS_child_unlock(container);
  
  return data;
}


void *ROWS_container_lookup_int(ROWS_container container,
				ROWS_child child)
{
   _p_ROWS_container_data   **priv;
  ROWS_container_child      *next;
  void                      *data = NULL;

  ROWS_widget_check_type(container, container_type, "ROWS_container");
  /* ROWS_widget_lock(container); */
  ROWS_child_lock(child);

  priv = (_p_ROWS_container_data **) container;
  next = (*priv)->children;

  while (next)
  {
    if (next->child == child)
      data = next->data;
    
    next = next->next;
  }

  ROWS_child_unlock(child);

  return data;
}


void ROWS_container_iter_int(ROWS_container container,
			     ROWS_container_iter function,
			     void *moredata)
{
  _p_ROWS_container_data   **priv;
  ROWS_container_child      *next;
  
  ROWS_widget_check_type(container, container_type, "ROWS_container");
  ROWS_child_lock(container);
  
  priv   = (_p_ROWS_container_data **) container;
  next = (*priv)->children;
  
  while (next)
  {
    (*function)(container,
		next->child,
		next->data,
		moredata);
    
    next = next->next;
  }
  
  ROWS_child_unlock(container);
}

