/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * mapping.c
 * ---------
 *
 * Implementation of dynamic character set conversions
 */

/*********************************************************************
 * 
 * VIW (Very Important Warning)
 * 
 * No functions used impulse_module_* must *EVER* call
 * 
 *  impulse_unistring_to_8bit
 *  impulse_unistring_from_8bit
 * 
 * Both of these functions make use of dynamically loaded modules,
 * and thus would have huge potential for infinite loops.....
 * 
 ********************************************************************/

#include "../../../config.h"

#include "impulse/util/mapping.h"
#include "impulse/util/hash.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"
#include "impulse/util/module.h"
#include "impulse/util/error.h"

#include <stdlib.h>

static impulse_hash encodings = NULL;
#if USE_THREADS
static impulse_mutex_static encoding_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif

struct __p_impulse_mapping_info
{
  impulse_mapping mapping;
  impulse_int refs;
  impulse_module module;
};

typedef struct __p_impulse_mapping_info *_p_impulse_mapping_info;


_p_impulse_mapping_info _p_impulse_mapping_create(impulse_unistring encoding)
{
  impulse_char *envpath = getenv("IMPULSE_MAPPING_PATH");
  impulse_unistring path = NULL;
  impulse_module module = NULL;
  impulse_unistring tosymstr = NULL;
  impulse_unistring fromsymstr = NULL;
  void *tosym = NULL;
  void *fromsym = NULL;
  _p_impulse_mapping_info info = NULL;
  
  if (envpath)
  {
    path = impulse_unistring_create_utf(envpath, 8);
    impulse_unistring_append_utf(path, "/", 8);
  }
  else
  {
    path = impulse_unistring_create_utf(PREFIX_DIR, 8);
    impulse_unistring_append_utf(path, "/lib/impulse/modules/mapping/", 8);
  }
  
  impulse_unistring_append(path, encoding);
  
  if (!(module = impulse_module_create(path)))
  {
    impulse_unistring_destroy(path);
    return NULL;
  }
  impulse_unistring_destroy(path);
  
  tosymstr = impulse_unistring_create_utf("to_8bit", 8);
  fromsymstr = impulse_unistring_create_utf("from_8bit", 8);
  if (!(tosym = impulse_module_symbol(module, tosymstr)) ||
      !(fromsym = impulse_module_symbol(module, fromsymstr)))
  {
    impulse_unistring_destroy(tosymstr);
    impulse_unistring_destroy(fromsymstr);
    impulse_module_destroy(module);
    return NULL;
  }
  impulse_unistring_destroy(tosymstr);
  impulse_unistring_destroy(fromsymstr);
  
  info = impulse_alloc(sizeof(struct __p_impulse_mapping_info));
  info->module = module;
  info->refs = 1;
  info->mapping = impulse_alloc(sizeof(struct _impulse_mapping));
  info->mapping->encoding = impulse_unistring_create(encoding);
  info->mapping->to = tosym;
  info->mapping->from = fromsym;
  
  return info;
}


impulse_mapping impulse_mapping_create(impulse_unistring encoding)
{
  _p_impulse_mapping_info info = NULL;

#if USE_THREADS
  impulse_mutex_lock(&encoding_mutex);
#endif
  if (!encodings)
    encodings = impulse_hash_create();
  
  if (!(info = impulse_hash_recover(encodings, encoding)))
  {
    info = _p_impulse_mapping_create(encoding);
    if (!info)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Could not load character encoding database", 8);
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, encoding);
      impulse_unistring_destroy(desc);
    }
    else
      impulse_hash_store(encodings, encoding, info);
  }
  else
    info->refs++;
  
#if USE_THREADS
  impulse_mutex_unlock(&encoding_mutex);
#endif
  return info->mapping;
}


void impulse_mapping_destroy(impulse_mapping map)
{
  _p_impulse_mapping_info info = NULL;

#if USE_THREADS
  impulse_mutex_lock(&encoding_mutex);
#endif
  if (!encodings)
    encodings = impulse_hash_create();
  
  if (!(info = impulse_hash_recover(encodings, map->encoding)))
  {
    impulse_unistring desc = impulse_unistring_create_utf("Encodings database gone all wierd", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, map->encoding);
    impulse_unistring_destroy(desc);
  }

  if (!(--(info->refs)))
  {
    impulse_hash_remove(encodings, info->mapping->encoding);
    
    impulse_unistring_destroy(info->mapping->encoding);
    impulse_free(info->mapping);
    impulse_module_destroy(info->module);
    impulse_free(info);
  }
#if USE_THREADS
  impulse_mutex_unlock(&encoding_mutex);
#endif
}


/* End of file */
