/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * unistring.c
 * -----------
 *
 * Functions to manipulate Unicode strings
 */

#include "../../../config.h"

#include "impulse/util/unistring.h"
#include "impulse/util/mapping.h"
#include "impulse/util/unidata.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>


impulse_unistring impulse_unistring_create_space(impulse_int capacity)
{
  impulse_unistring str;
  str = impulse_alloc(sizeof(struct _impulse_unistring));
  
  str->capacity = capacity;
  str->length = 0;
  if (capacity)
  {
    str->data = impulse_alloc(sizeof(impulse_unichar) * (capacity+1));
    str->data[0] = 0;
  }
  
  return str;
}


impulse_unistring impulse_unistring_create(impulse_unistring src)
{
  impulse_unistring str = impulse_alloc(sizeof(struct _impulse_unistring));
  str->capacity = src->capacity;
  str->length = src->length;
  str->data = impulse_alloc((str->capacity+1)*sizeof(impulse_unichar));  
  memcpy(str->data, src->data, str->length*sizeof(impulse_unichar));
  str->data[str->length] = 0;
  
  return str;
}


impulse_unistring impulse_unistring_create_utf(impulse_char *src,
					       impulse_int bits)
{
  impulse_unistring str = impulse_alloc(sizeof(struct _impulse_unistring));
  str->capacity = 0;
  str->length = 0;
  str->data = NULL;
  
  impulse_unistring_from_utf(str, src, bits);
  
  return str;
}


impulse_unistring impulse_unistring_create_8bit(impulse_char *src,
						impulse_unistring map)
{
  impulse_unistring str = impulse_alloc(sizeof(struct _impulse_unistring));
  str->capacity = 0;
  str->length = 0;
  str->data = NULL;
  
  impulse_unistring_from_8bit(str, src, map);
  
  return str;
}


void impulse_unistring_destroy(impulse_unistring str)
{
  if (str->data)
    impulse_free(str->data);
  impulse_free(str);
}

/* Conversion functions */

/* UTF-7 information */
static const char base64[] =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
/* 1 = Set D, 2 = Set O, 4 = (Rule 3) */
static const impulse_byte direct[] =
{
  0,0,0,0,0,0,0,0,0,4,4,0,0,4,0,0, /* 0-15    */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 16-31   */
  4,2,2,2,2,2,2,1,1,1,2,0,1,1,1,1, /* 32-47   */
  1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,1, /* 48-63   */
  2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 64-79   */
  1,1,1,1,1,1,1,1,1,1,1,2,0,2,2,2, /* 80-95   */
  2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 96-111  */
  1,1,1,1,1,1,1,1,1,1,1,2,2,2,0,0  /* 112-127 */
};

static inline void impulse_unistring_base64(int  code,
					    impulse_char  *result,
					    impulse_int    bits)
{
  int    bit, codename;
  impulse_uint work;
  int    chars;
  
  work = code;
    
  chars = 4;
  if (bits<24)
    {
      /* Pad to character boundary */
      if ((bits%6) != 0)
	{
	  chars = (bits/6) + 1;
	}
      else
	{
	  chars = bits/6;
	}
    }

  for (bit=0; bit<chars; bit++)
    {
      codename = (work&0xfc0000) >> 18;
      result[bit] = base64[codename];
      work <<= 6;
    }
}

/*
 * Options should be 0 for default behaviour in all encodings
 *
 * Note: Options is used in UTF-7 encoding, and specifies which of the 
 * various types of outputs. It is a flag value, and defines which
 * characters are encoded directly. Depending on the dataset, you may
 * find altering it helpful.
 *
 * Flag values:
 * 1 - encode set D characters in Base64 (alphanumeric & some punctuation)
 * 2 - encode set O characters in Base64 (more punctuation. Setting
 *       this makes data compatible with mail gateways)
 * 4 - encode rule 3 characters in Base64 (whitespace. Makes data
 *       compatible with MIME)
 *
 * ie - 0 gives as much plaintext as possible, 7 gives all Base64
 */

char *impulse_unistring_to_utf(impulse_unistring string, 
			       impulse_int bitlength, 
			       impulse_int options)
{
  char *result = NULL;
  int   reslen = 0;
  
  switch (bitlength)
    {
    case 8: /* UTF-8 */
      {
	int pos;

	for (pos=0; pos<string->length; pos++)
	  {
	    /*
	     * FIXME:we could speed things up by only checking for
	     * surrogates in form 3 (zzzzyyyyyyxxxxxx)
	     */
	    if (string->data[pos] >= UNICODE_SURROGATEHIGH_START &&
		string->data[pos] <= UNICODE_SURROGATEHIGH_END)
	      {
		int lowpos;

		/* FIXME: we could encode orphan high surrogates... */
		
		lowpos = pos+1;
		if (lowpos < string->length-1)
		  {
		    if (string->data[lowpos] >= UNICODE_SURROGATELOW_START &&
			string->data[lowpos] <= UNICODE_SURROGATELOW_END)
		      {
			impulse_unichar a,b;
			char    w,z,y,x;
			int     start;

			a = string->data[pos];
			b = string->data[lowpos];

			w = (a>>6)&0x0f;
			z = (a>>2)&0x0f;
			y = ((a&0x03)<<4)|((b>>6)&0x0f);
			x = b&0x3f;

			w++;

			start = reslen;
			reslen+=4;
			result = impulse_realloc(result,
						 reslen*sizeof(char));
			result[start]   = (w>>2)     | 0xf0;
			result[start+1] = ((w<<4)|z) | 0x80;
			result[start+2] = y          | 0x80;
			result[start+3] = x          | 0x80;
			
			pos = lowpos;
		      }
		  }
	      }
	    else
	      {
		/* Three other forms */
		if ((string->data[pos]&0xff80) == 0)
		  {
		    int start;

		    /* 000000000xxxxxxx */
		    start = reslen;
		    reslen++;
		    result = impulse_realloc(result,
					     reslen*sizeof(char));
		    result[start] = string->data[pos];
		  }
		else if ((string->data[pos]&0xf800) == 0)
		  {
		    int start;
		    
		    /* 00000yyyyyxxxxxx */
		    start = reslen;
		    reslen+=2;
		    result = impulse_realloc(result,
					     reslen*sizeof(char));
		    result[start]   = (string->data[pos]>>6)   | 0xc0;
		    result[start+1] = (string->data[pos]&0x3f) | 0x80;
		  }
		else
		  {
		    int start;
		    
		    /* zzzzyyyyyyxxxxxx */
		    start = reslen;
		    reslen+=3;
		    result = impulse_realloc(result,
					     reslen*sizeof(char));
		    result[start]   = (string->data[pos]>>12)       | 0xe0;
		    result[start+1] = ((string->data[pos]>>6)&0x3f) | 0x80;
		    result[start+2] = (string->data[pos]&0x3f)      | 0x80;
		  }
	      }
	  }
	reslen++;
	result = impulse_realloc(result,
				 reslen*sizeof(char));
	result[reslen-1] = 0;
      }
      break;

    case 7: /* UTF-7 */
      {
	enum { DIRECT, BASE64 } lastenc = DIRECT;
	int   lastcode = 0;
	int     codepos  = 0;
	int     pos;
	int     is_direct;
	int   c;

	options = (~options)&7;

	for (pos=0; pos<string->length; pos++)
	  {
	    c = string->data[pos];
	    
	    is_direct = 0;
	    if (c < 128)
	      if ((direct[c]&options) != 0)
		{
		  is_direct = 1;
		}

	    if (is_direct == 1)
	      {
		if (codepos != 0 && lastenc == BASE64)
		  {
		    int start;
		    
		    start = reslen;
		    reslen+=(codepos/6)+1;
		    result = impulse_realloc(result, reslen*sizeof(char));
		    impulse_unistring_base64(lastcode, &result[start], codepos);
		    
		    codepos  = 0;
		    lastcode = 0;
		  }

		if (lastenc == BASE64)
		  {
		    reslen++;
		    result = impulse_realloc(result, reslen*sizeof(char));
		    result[reslen-1] = '-';
		    lastenc = DIRECT;
		  }

		reslen++;
		result = impulse_realloc(result, reslen*sizeof(char));
		result[reslen-1] = c;
	      }
	    else
	      {
		if (lastenc == DIRECT)
		  {
		    reslen++;
		    result = impulse_realloc(result, reslen*sizeof(char));
		    result[reslen-1] = '+';
		    codepos  = 0;
		    lastcode = 0;
		    lastenc  = BASE64;
		  }

		switch (codepos)
		  {
		  case 16:
		    {
		      int start;
		      
		      lastcode = lastcode | (((int)(c>>8)));
		      
		      start = reslen;
		      reslen+=4;
		      result = impulse_realloc(result, reslen*sizeof(char));
		      impulse_unistring_base64(lastcode,
					     &result[start], 
					     24);
		      
		      codepos = 8;
		      lastcode = (c&0xff)<<16;
		    }
		    break;
		    
		  case 8:
		    {
		      int start;

		      lastcode = lastcode | c;

		      start = reslen;
		      reslen+=4;
		      result = impulse_realloc(result, reslen*sizeof(char));
		      impulse_unistring_base64(lastcode,
					     &result[start], 
					     24);
		      
		      codepos = 0;
		      lastcode = 0;
		    }
		    break;

		  case 0:
		    {
		      codepos = 16;
		      lastcode = c<<8;
		    }
		  }
	      }
	  }

	if (lastenc == BASE64)
	  {
	    int start;
	    
	    start = reslen;
	    switch (codepos)
	      {
	      case 0:
		reslen+=1;
		break;
	      case 8:
		reslen+=3;
		break;
	      case 16:
		reslen+=4;
		break;
	      case 24:
		reslen+=5;
		break;
	      }
	    result = impulse_realloc(result, reslen*sizeof(char));
	    impulse_unistring_base64(lastcode, &result[start], codepos);
	    result[reslen-1] = '-';
	  }

	reslen+=1;
	result = impulse_realloc(result, reslen*sizeof(char));
	result[reslen-1] = 0;
      }
      break;
      
    default:
      {
	char *yahboo = "Unknown bitlength";
	result = impulse_alloc(strlen(yahboo)+1);
	strcpy(result, yahboo);
      }
      break;
    }

  return result;
}


static const impulse_byte base64_trans[128] =
{
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255, /* 0-15    */
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255, /* 16-31   */
  255,255,255,255,255,255,255,255,255,255,255, 62,255,255,255, 63, /* 32-47   */
   52, 53, 54, 55, 56, 57, 58, 59, 60, 61,255,255,255,255,255,255, /* 48-63   */
  255,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, /* 64-79   */
   15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,255,255,255,255,255, /* 80-95   */
  255, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, /* 96-111  */
   41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51,255,255,255,255,255  /* 112-127 */
};  

/*
 * FIXME: UTF-8 decoding may be incomplete. Not having read ISO 10646
 * I wouldn't know, but I suspect it's possible to use a 32-bit
 * encoding to represent a valid Unicode character.
 */
void impulse_unistring_from_utf(impulse_unistring string, 
				impulse_char *chunk, 
				impulse_int bitlength)
{
  int i;
  impulse_free(string->data);
  string->data = NULL;
  string->length = string->capacity = 0;

  string->length = strlen(chunk);
  string->capacity = string->length;
  string->data = impulse_alloc(sizeof(impulse_unichar) * (string->length + 1));
  for (i = 0 ; i <= string->length ; i++)
    string->data[i] = (impulse_unichar)chunk[i];
  return;
  
  switch(bitlength)
    {
    case 8:
      {
	int pos, len, spos;
	char c;
	impulse_unichar out;
	
	pos = 0;
	len = strlen(chunk);

	for (pos=0; pos<len; pos++)
	  {
	    c = chunk[pos];

	    spos = string->length;
	    string->length++;
	    string->data = impulse_realloc(string->data,
					   (string->length+1)*sizeof(impulse_unichar));

	    if ((c&0x80) == 0)
	      {
		/* Direct mapping */
		string->data[spos] = (impulse_unichar) c;
	      }
	    else
	      {
		if ((c&0xe0) == 0xc0 && (pos+1)<len)
		  {
		    /* 00000yyyyyxxxxxx */
		    if ((chunk[pos+1]&0xc0) != 0x80)
		      {
			impulse_unistring desc = impulse_unistring_create_utf("Some wierd juju while converting from utf", 8);
			impulse_error_raise(IMPULSE_ERROR_WARNING, desc, 0);
			impulse_unistring_destroy(desc);
		      }

		    out =
		      (((impulse_unichar)(c&0x1f))             << 6) |
		      (((impulse_unichar)(chunk[pos+1]&0x3f)) << 0);

		    string->data[spos] = out;
		    
		    pos++;
		  }
		else if ((c&0xf0) == 0xe0 && (pos+2)<len)
		  {
		    /* zzzzyyyyyyxxxxxx */
		    if ((chunk[pos+1]&0xc0) != 0x80 ||
			(chunk[pos+2]&0xc0) != 0x80)
		      {
                        impulse_unistring desc = impulse_unistring_create_utf("Some wierd juju while converting from utf", 8);
			impulse_error_raise(IMPULSE_ERROR_WARNING, desc, 0);
			impulse_unistring_destroy(desc);
		      }

		    out =
		      (((impulse_unichar)(c&0x0f))             << 12) |
		      (((impulse_unichar)(chunk[pos+1]&0x3f)) << 6)  |
		      (((impulse_unichar)(chunk[pos+2]&0x3f)) << 0);

		    string->data[spos] = out;

		    pos+=2;
		  }
		else if ((c&0xf8) == 0xf0 && (pos+3)<len)
		  {
		    impulse_unichar out2;
		    int     w;
		    
		    /* Surrogate character */
		    if ((chunk[pos+1]&0xc0) != 0x80 ||
			(chunk[pos+2]&0xc0) != 0x80 ||
			(chunk[pos+3]&0xc0) != 0x80)
		      {
                        impulse_unistring desc = impulse_unistring_create_utf("Some wierd juju while converting from utf", 8);
			impulse_error_raise(IMPULSE_ERROR_WARNING, desc, 0);
			impulse_unistring_destroy(desc);
		      }

		    string->length++;
		    string->data = impulse_realloc(string->data,
						   (string->length+1)*
						   sizeof(impulse_unichar));

		    w =
		      (((impulse_unichar)(chunk[pos]&0x07))   << 2) |
		      (((impulse_unichar)(chunk[pos+1]&0x30)) >> 4);
		    w--;

		    out = 0xd800 |
		      w<<6       |
		      (((impulse_unichar)(chunk[pos+1]&0x0f)) << 2) |
		      (((impulse_unichar)(chunk[pos+2]&0x30)) >> 4);
		    out2 = 0xdc00 |
		      (((impulse_unichar)(chunk[pos+2]&0x0f)) << 6) |
		      (((impulse_unichar)(chunk[pos+3]&0x3f)) << 0);

		    string->data[spos]   = out;
		    string->data[spos+1] = out2;
		    
		    pos+=3;
		  }
		else
		  {
		    /* Error */
		    string->length = 0;
		    string->capacity = 0;
		    if (string->data != NULL)
		      impulse_free(string->data);
		    string->data = NULL;
		    return;
		  }
	      }
	  }
      }
      break;

    case 7:
      {
	int pos, len;
	
	string->length = 0;
	pos = 0;
	len = strlen(chunk);

	for (pos=0; pos<len; pos++)
	  {
	    if (chunk[pos] == '+')
	      {
		int epos, base_len;
		int unicode_len, byte_len;

		epos = pos+1;
		while (epos<len &&
		       base64_trans[(int) chunk[epos]] != 255)
		  epos++;

		base_len    = epos-(pos+1);
		byte_len    = (base_len*3)/4;
		unicode_len = (base_len*3)/8;

		if (unicode_len > 0)
		  {
		    char *out;
		    int temp;
		    int x, y, outpos, shift;
		    int spos;

		    spos = string->length;
		    string->length += unicode_len;
		    string->data =  impulse_realloc(string->data,
						    (string->length+1)*
						    sizeof(impulse_unichar));

		    shift  = 18;
		    out    = impulse_alloc((byte_len+1)*sizeof(char));
		    outpos = 0;
		    temp   = 0;
		    
		    for (x=pos+1; x<epos; x++)
		      {
			temp |= ((int) base64_trans[(int) chunk[x]])<<shift;

			shift -= 6;
			if (shift < 0)
			  {
			    for (y=2; y>=0; y--)
			      {
				out[outpos+y] = (char) (temp&0xff);
				temp >>= 8;
			      }
			    outpos+=3;
			    
			    temp  = 0;
			    shift = 18;
			  }
		      }

		    if (shift < 18)
		      {
			int bytesleft;

			bytesleft = (18-shift)/8;
			for (y=0; y<bytesleft; y++)
			  {
			    out[outpos] = (char)((temp&0xff0000) >> 16);
			    outpos++;
			    temp<<=8;
			  }
		      }

		    for (x=0; x<unicode_len; x++)
		      {
			string->data[spos+x] =
			  (((impulse_unichar)out[x*2])<<8) |
			  (((impulse_unichar)out[(x*2)+1]));
		      }
		  }
		else
		  {
		    string->length++;
		    string->data = impulse_realloc(string->data,
						   (string->length+1)*
						   sizeof(impulse_unichar));
		    string->data[string->length-1] = (impulse_unichar) '+';
		  }

		if (epos <= (pos+1))
		  epos++;
		pos = epos;
	      }
	    else
	      {
		/*
		 * Maybe we should check that the character is indeed
		 * a set O, D or rule 3 one
		 */
		string->length++;
		string->data = impulse_realloc(string->data,
					       (string->length+1)*
					       sizeof(impulse_unichar));
		string->data[string->length-1] = (impulse_unichar) chunk[pos];
	      }
	  }
      }
      break;
    }
  string->capacity = string->length;
  string->data[string->length] = 0;
}


char *impulse_unistring_to_8bit(impulse_unistring string, 
				impulse_unistring mapname)
{
  impulse_mapping map = impulse_mapping_create(mapname);
  impulse_mapping_to_8bit m;
  char                   *result = NULL;
  int                     reslen = 0;
  int                     pos;
  
  m = map->to;

  for (pos=0; pos<string->length; pos++)
    {
      reslen++;
      result = impulse_realloc(result, reslen*sizeof(char));
      
      /* FIXME: No provision is made for surrogates, yet */
      if (string->data[pos] >= UNICODE_SURROGATEHIGH_START &&
	  string->data[pos] <= UNICODE_SURROGATEHIGH_END)
	{
	  int lowpos;

	  result[reslen-1] = '?';

	  lowpos = pos+1;
	  if (lowpos < string->length-1)
	    {
	      if (string->data[lowpos] >= UNICODE_SURROGATELOW_START &&
		  string->data[lowpos] <= UNICODE_SURROGATELOW_END)
		{
		  pos = lowpos;
		}
	    }
	}
      else
	{
	  int d1, d2;

	  d1 = (string->data[pos]>>8)&0xff;
	  d2 = string->data[pos]&0xff;

	  switch (m->mapping[d1][d2].type)
	    {
	    case UNICODE_MAP_UNMAPPED:
	      /* FIXME: deal with this in a nicer way */
	      result[reslen-1] = '?';
	      break;
	      
	    case UNICODE_MAP_SINGLE:
	      result[reslen-1] = m->mapping[d1][d2].mapchar;
	      break;

	    case UNICODE_MAP_MULTIPLE:
	      {
		int startpos;

		startpos = reslen-1;
		reslen = startpos + m->mapping[d1][d2].mapstr.length;
		result = impulse_realloc(result, reslen*sizeof(char));

		memcpy(&result[startpos],
		       m->mapping[d1][d2].mapstr.string,
		       m->mapping[d1][d2].mapstr.length*sizeof(char));
	      }
	      break;
	    }
	}
    }

  reslen++;
  result = impulse_realloc(result, reslen*sizeof(char));
  result[reslen-1] = 0;
  impulse_mapping_destroy(map);
  return result;
}


void impulse_unistring_from_8bit(impulse_unistring string, 
				 impulse_char *chunk, 
				 impulse_unistring mapname)
{
  impulse_mapping map = impulse_mapping_create(mapname);
  impulse_mapping_from_8bit m;
  int pos = 0;

  impulse_free(string->data);
  string->data = NULL;
  string->capacity = string->length = 0;

  m = map->from;
  
  while (chunk[pos] != 0)
    {
      char c;

      c = chunk[pos];
      
      string->length++;
      string->data = impulse_realloc(string->data,
				   (string->length+1)*sizeof(impulse_unichar));

      switch (m->mapping[(int)c].type)
	{
	case UNICODE_MAP_UNMAPPED:
	  /* Replacement character U+FFFD, used for unknown scripts */
	  string->data[string->length-1] = 0xfffd;
	  break;
	case UNICODE_MAP_SINGLE:
	  string->data[string->length-1] = m->mapping[(int)c].mapchar;
	  break;
	case UNICODE_MAP_MULTIPLE:
	  {
	    int startpos;

	    startpos = string->length - 1;
	    string->length = startpos + m->mapping[(int)c].mapstr.length;
	    string->data = impulse_realloc(string->data,
					 (string->length+1)*sizeof(impulse_unichar));
	    memcpy(&string->data[startpos],
		   m->mapping[(int)c].mapstr.string,
		   m->mapping[(int)c].mapstr.length*sizeof(impulse_unichar));
	  }
	  break;
	}
      
      pos++;
    }
  string->capacity = string->length;
  string->data[string->length] = 0;
  
  impulse_mapping_destroy(map);
}


void impulse_unistring_decompose(impulse_unistring string)
{
  int pos;

  for (pos=0; pos<string->length; pos++)
    {
      impulse_byte r1, r2;

      r1 = (impulse_byte) ((impulse_ushort)(string->data[pos]>>8));
      r2 = (impulse_byte) ((impulse_ushort)(string->data[pos]&0xff));  
      do
	{
	  int len;

	  len = impulse_unidata[r1][r2].decomposition.length;
	  
	  if (len > 1 &&
	      impulse_unidata[r1][r2].decomposition.formatting == 0)
	    {
	      impulse_unichar *newstr;
	      int      newpos;

	      newstr = impulse_realloc(string->data,
				       (string->length + (len-1)+1) *
				       sizeof(impulse_unichar));

	      memmove(&(newstr[pos+len]),
		      &(newstr[pos+1]),
		      (string->length - pos - 1)*sizeof(impulse_unichar));
	      
	      for (newpos=0; newpos<len; newpos++)
		{
		  newstr[pos+newpos] =
		    impulse_unidata[r1][r2].decomposition.decomposition[newpos];
		}
	      
	      string->data  = newstr;
	      string->length += len-1;
	    }

	  r1 = (impulse_byte) ((impulse_ushort)(string->data[pos]>>8));
	  r2 = (impulse_byte) ((impulse_ushort)(string->data[pos]&0xff));
	} while (impulse_unidata[r1][r2].decomposition.length     >  1 &&
		 impulse_unidata[r1][r2].decomposition.formatting == 0);
    }

  string->capacity = string->length;
}


void impulse_unistring_compose(impulse_unistring string)
{
  impulse_unistring result;
  impulse_unichar   cur, lookahead;
  int pos;

  if (string->length <= 1)
    return;

  result = impulse_unistring_create_space(0);

  lookahead = string->data[0];
  for (pos=1; pos<string->length; pos++)
    {
      void **array;
      impulse_unidata_composition *comp;
      int x;
      impulse_uint entry;

      cur = lookahead;
      lookahead = string->data[pos];

      comp = NULL;
      array = (void **) impulse_unidata_comp_table;

      entry = (cur<<16)|lookahead;
      
      for (x=0;x<15;x++)
	{
	  int p;

	  p = entry&3;
	  entry >>= 2;

	  if (array[p] == NULL)
	    {
	      array = NULL;
	      x = 17;
	    }
	  else
	    {
	      array = array[p];
	    }
	}

      if (array != NULL)
	{
	  if (array[entry] != NULL)
	    {
	      /* We have a match */
	      comp = (impulse_unidata_composition *) &array[entry];

	      if (comp->flags == 1)
		{
		  result->length++;
		  result->data = impulse_realloc(result->data,
						 (result->length+1)*sizeof(impulse_unichar));
		  result->data[result->length-1] = comp->composition;
		  
		  pos++;
		  if (pos < string->length)
		    {
		      lookahead = string->data[pos];
		    }
		}
	      else
		{
		  if ((comp->flags&2) != 0)
		    {
		      int x;
		      int match_no, match_length;

		      match_no = -1;
		      match_length = -1;
		      
		      /* Try and find the longest complete match */
		      /* FIXME: there are faster ways of doing this...*/
		      for (x=0; x<comp->number_multis; x++)
			{
			  if ((pos+comp->multi[x].length-1) <=
			      string->length &&
			      comp->multi[x].length > match_length)
			    { 
			      if (memcmp(comp->multi[x].string,
					 &string->data[pos-1],
					 comp->multi[x].length*sizeof(impulse_unichar)
					 ) == 0)
				{
				  match_no     = x;
				  match_length = comp->multi[x].length;
				}
			    }
			}

		      if (match_no != -1)
			{
			  result->length++;
			  result->data = impulse_realloc(result->data,
							 (result->length+1)*
							 sizeof(impulse_unichar));
			  result->data[result->length-1] =
			    comp->multi_compo[match_no];
			  
			  pos += match_length-1;
			  if (pos < string->length)
			    {
			      lookahead = string->data[pos];
			    }
			}
		      else
			{
			  if ((comp->flags&1)!=0)
			    {
			      result->length++;
			      result->data = impulse_realloc(result->data,
							     (result->length+1)*
							     sizeof(impulse_unichar));
			      result->data[result->length-1] =
				comp->composition;
			      
			      pos++;
			      if (pos < string->length)
				{
				  lookahead = string->data[pos];
				}			      
			    }
			  else
			    {
			      result->length++;
			      result->data = impulse_realloc(result->data,
							     (result->length+1)*
							     sizeof(impulse_unichar));
			      result->data[result->length-1] = cur;
			    }
			}
		    }
		  else
		    {
		      result->length++;
		      result->data = impulse_realloc(result->data,
						     (result->length+1)*
						     sizeof(impulse_unichar));
		      result->data[result->length-1] = cur;
		    }
		}
	    }
	}
      else
	{
	  result->length++;
	  result->data = impulse_realloc(result->data,
					 (result->length+1)*sizeof(impulse_unichar));
	  result->data[result->length-1] = cur;
	}
    }

  if (pos == string->length)
    {
      result->length++;
      result->data = impulse_realloc(result->data,
				     (result->length+1)*sizeof(impulse_unichar));
      result->data[result->length-1] = lookahead;      
    }
  
  impulse_unistring_assign(string, result);
  impulse_unistring_destroy(result);
}


void impulse_unistring_bidirection(impulse_unistring string)
{
}



/* Manipulation functions */


void impulse_unistring_resize(impulse_unistring str,
			      impulse_int len)
{
  if (len < str->length)
    str->length = len;
  str->capacity = len;
  str->data = impulse_realloc(str->data, (len+1)*sizeof(impulse_unichar));
  str->data[str->length] = 0;
}


void impulse_unistring_trim(impulse_unistring str)
{
  /* Remove whitespace from either end of string */
}


void impulse_unistring_substring(impulse_unistring dst,
				 impulse_unistring src,
				 impulse_int start,
				 impulse_int end)
{
  int length = 0;
  if (!src->length)
    return;
  if (start > end)
    start = end = 0;
/*  {
    start ^= end;
    end ^= start;
    start ^= end;
  }*/
  if (start > src->length)
    start = src->length - 1;
  if (end > src->length)
    end = src->length - 1;
  
  length = end - start;
  if (!length)
  {
    dst->length = 0;
    dst->capacity = 0;
    if (dst->data)
      impulse_free(dst->data);
    dst->data = NULL;
  }
  else
  {
    if (dst->capacity < length)
    {
      if (dst->data)
	impulse_free(dst->data);
      dst->capacity = length;
      dst->data = impulse_alloc((dst->capacity+1)*sizeof(impulse_unichar));
    }
    memmove(dst->data, src->data+start, (length)*sizeof(impulse_unichar));
    dst->length = length;
    dst->data[dst->length] = 0;
  }
}


void impulse_unistring_assign(impulse_unistring str,
			      impulse_unistring chunk)
{
  str->length = chunk->length;
  if (str->capacity < str->length)
  {
    impulse_free(str->data);
    str->capacity = str->length;
    str->data = impulse_alloc((str->capacity+1)*sizeof(impulse_unichar));
  }
  
  memcpy(str->data, chunk->data, str->length*sizeof(impulse_unichar));
  str->data[str->length] = 0;
}


void impulse_unistring_append(impulse_unistring str,
			      impulse_unistring chunk)
{
  impulse_int length = str->length + chunk->length;
  if (str->capacity < length)
  {
    str->capacity = length;
    str->data = impulse_realloc(str->data, (str->capacity+1)*sizeof(impulse_unichar));
  }
  memcpy(str->data+str->length,chunk->data,chunk->length*sizeof(impulse_unichar));
  str->length = length;
  str->data[str->length] = 0;
}


void impulse_unistring_prepend(impulse_unistring str,
			       impulse_unistring chunk)
{
  impulse_int length = str->length + chunk->length;
  if (str->capacity < length)
  {
    str->capacity = length;
    str->data = impulse_realloc(str->data, (str->capacity+1)*sizeof(impulse_unichar));
  }
  memmove(str->data+chunk->length, str->data,str->length*sizeof(impulse_unichar));
  memcpy(str->data, chunk->data, chunk->length*sizeof(impulse_unichar));
  str->length = length;
  str->data[str->length] = 0;
}


void impulse_unistring_insert(impulse_unistring str,
			      impulse_unistring chunk,
			      impulse_int offset)
{
  impulse_int length = str->length + chunk->length;
  if (offset < 0)
    offset = 0;
  if (offset > str->length)
    offset = str->length;
  if (str->capacity < length)
  {
    str->capacity = length;
    str->data = impulse_realloc(str->data, (str->capacity+1)*sizeof(impulse_unichar));
  }
  memmove(str->data+offset+chunk->length, str->data+offset, (str->length-offset)*sizeof(impulse_unichar));
  memcpy(str->data+offset, chunk->data, chunk->length*sizeof(impulse_unichar));
  str->length = length;
  str->data[str->length] = 0;
}


void impulse_unistring_assign_utf(impulse_unistring str,
				  impulse_char *chunk,
				  impulse_int bits)
{
  impulse_unistring uchunk = impulse_unistring_create_utf(chunk, bits);
  impulse_unistring_assign(str, uchunk);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_append_utf(impulse_unistring str,
				  impulse_char *chunk,
				  impulse_int bits)
{
  impulse_unistring uchunk = impulse_unistring_create_utf(chunk, bits);
  impulse_unistring_append(str, uchunk);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_prepend_utf(impulse_unistring str,
				   impulse_char *chunk,
				   impulse_int bits)
{
  impulse_unistring uchunk = impulse_unistring_create_utf(chunk, bits);
  impulse_unistring_prepend(str, uchunk);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_insert_utf(impulse_unistring str,
				  impulse_char *chunk,
				  impulse_int bits,
				  impulse_int offset)
{
  impulse_unistring uchunk = impulse_unistring_create_utf(chunk, bits);
  impulse_unistring_insert(str, uchunk, offset);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_assign_8bit(impulse_unistring str,
				   impulse_char *chunk,
				   impulse_unistring map)
{
  impulse_unistring uchunk = impulse_unistring_create_8bit(chunk, map);
  impulse_unistring_assign(str, uchunk);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_append_8bit(impulse_unistring str,
				   impulse_char *chunk,
				   impulse_unistring map)
{
  impulse_unistring uchunk = impulse_unistring_create_8bit(chunk, map);
  impulse_unistring_append(str, uchunk);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_prepend_8bit(impulse_unistring str,
				    impulse_char *chunk,
				    impulse_unistring map)
{
  impulse_unistring uchunk = impulse_unistring_create_8bit(chunk, map);
  impulse_unistring_prepend(str, uchunk);
  impulse_unistring_destroy(uchunk);
}


void impulse_unistring_insert_8bit(impulse_unistring str,
				   impulse_char *chunk,
				   impulse_unistring map,
				   impulse_int offset)
{
  impulse_unistring uchunk = impulse_unistring_create_8bit(chunk, map);
  impulse_unistring_insert(str, uchunk, offset);
  impulse_unistring_destroy(uchunk);
}


/* Search functions */


/*
 * Fix me: The functions which follow for finding a string 
 * within a string all use the naive searching algorithm.
 * This is O(m*n) in worst case.
 * We really want something clever such as the Knuth-Morris
 * Pratt algorithm which is O(m+n) is worst case - quite an
 * improvement really.
 */

impulse_int impulse_unistring_find(impulse_unistring str,
				   impulse_unistring chunk)
{
  impulse_int i;
  for (i = 0 ; i < str->length ; i++)
  {
    impulse_unichar *s1 = str->data + i, *s2 = chunk->data;
    impulse_bool match = 1;
    while (*s1 && *s2)
    {
      if (*s1 != *s2)
      {
        match = 0;
        break;
      }
      s1++;
      s2++;
    }
    
    if (match)
      return i;
  }

  return -1;
}


impulse_int impulse_unistring_rfind(impulse_unistring str,
				    impulse_unistring chunk)
{
  impulse_int i;
  for (i = str->length -1 ; i >= 0 ; i--)
  {
    impulse_unichar *s1 = str->data + i, *s2 = chunk->data;
    impulse_bool match = 1;
    while (*s1 && *s2)
    {
      if (*s1 != *s2)
      {
        match = 0;
        break;
      }
      s1++;
      s2++;
    }
    if (match)
      return i;
  }
  return -1;
}


impulse_int impulse_unistring_find_from(impulse_unistring str,
					impulse_unistring chunk,
					impulse_int offset)
{
  impulse_int i;
  if (offset < 0)
    return -1;
  for (i = offset ; i < str->length ; i++)
  {
    impulse_unichar *s1 = str->data + i, *s2 = chunk->data;
    impulse_bool match = 1;
    while (*s1 && *s2)
    {
      if (*s1 != *s2)
      {
        match = 0;
        break;
      }
      s1++;
      s2++;
    }
    
    if (match)
      return i;
  }

  return -1;
}


impulse_int impulse_unistring_rfind_from(impulse_unistring str,
					 impulse_unistring chunk,
					 impulse_int offset)
{
  impulse_int i;
  if (offset < 0)
    return -1;
  if (offset > (str->length -1))
    return -1;
  for (i = offset ; i >= 0 ; i--)
  {
    impulse_unichar *s1 = str->data + i, *s2 = chunk->data;
    impulse_bool match = 1;
    while (*s1 && *s2)
    {
      if (*s1 != *s2)
      {
        match = 0;
        break;
      }
      s1++;
      s2++;
    }
    if (match)
      return i;
  }
  return -1;
}


impulse_int impulse_unistring_find_chr(impulse_unistring str,
				       impulse_unichar c)
{
  impulse_unichar *start = str->data;

  while (*start)
    if (*start++ == c)
      return start - str->data - 1;
  
  return -1;
}


impulse_int impulse_unistring_rfind_chr(impulse_unistring str,
					impulse_unichar c)
{
  impulse_unichar *end = str->data+str->length-1;
  
  while (end && (end != str->data))
    if (*end-- == c)
      return end - str->data + 1;
  
  return -1;
}


impulse_int impulse_unistring_find_from_chr(impulse_unistring str,
					    impulse_unichar c,
					    impulse_int offset)
{
  impulse_unichar *start = str->data;

  if (offset > (str->length-1))
    return -1;
  if (offset < 0)
    return -1;
  start += offset;
  
  while (*start)
    if (*start++ == c)
      return start - str->data - 1;
  
  return -1;
}


impulse_int impulse_unistring_rfind_from_chr(impulse_unistring str,
					     impulse_unichar c,
					     impulse_int offset)
{
  impulse_unichar *end = str->data;
  if (offset > (str->length-1))
    return -1;
  if (offset < 0)
    return -1;
  end += offset;
  
  while (end && (end != str->data))
    if (*end-- == c)
      return end - str->data + 1;
  
  return 0;
}


/* Comparison functions */

impulse_bool impulse_unistring_equal(impulse_unistring str1,
				     impulse_unistring str2)
{
  impulse_unichar *s1 = str1->data, *s2 = str2->data;
  if (str1->length != str2->length)
    return 0;
  
  while (*s1)
    if (*s1++ != *s2++)
      return 0;

  return 1;
}


impulse_int impulse_unistring_compare(impulse_unistring str1,
				      impulse_unistring str2)
{
  return 0;
}


impulse_int impulse_unistring_compare_case(impulse_unistring str1,
					   impulse_unistring str2)
{
  return 0;
}


impulse_int impulse_unistring_ncompare(impulse_unistring str1,
				       impulse_unistring str2,
				       impulse_int count)
{
  return 0;
}


impulse_int impulse_unistring_ncompare_case(impulse_unistring str1,
					    impulse_unistring str2,
					    impulse_int count)
{
  return 0;
}



impulse_int impulse_unistring_starts(impulse_unistring str,
				     impulse_unistring chunk)
{
  impulse_unichar *dst = str->data, *src = chunk->data;
  if (str->length < chunk->length)
    return 0;

  while (*src)
    if (*dst++ != *src++)
      return 0;
  return 1;
}


impulse_int impulse_unistring_ends(impulse_unistring str,
				   impulse_unistring chunk)
{
  impulse_unichar *dst = str->data+(str->length-chunk->length), *src = chunk->data;
  if (str->length < chunk->length)
    return 0;

  while (*src)
    if (*dst++ != *src++)
      return 0;
  return 1;
}


/* End of file */
